
`include "config.sv"

module analyzer_clump(clk, reset, inputs, data_out_val, data_out, hps_read_addr, hps_next_addr, hps_can_read, block_configs, trig_configs, trig_conds, triggered, triggered_block_id, triggered_address, data_write_addr);


input  logic 		 							clk, reset;
input  logic [15:0] 							inputs;

output logic		 							data_out_val;
output logic [DPR_WORD_SIZE-1:0]			data_out;

// Generate Connections
input  logic [1:0]							block_configs	[0:3];
input  logic [1:0] 							trig_configs 	[0:3];
input  logic [TRIG_COND_BITWIDTH-1:0]	trig_conds	 	     ;

		 logic [DPR_WORD_SIZE-1:0]			block_data_out [0:3];
		 logic [DPR_ADDR_SIZE-1:0]			block_addr		[0:3];
		 logic [DPR_ADDR_SIZE-1:0]			block_next_addr[0:3];

// HPS SRAM
input  logic [DPR_ADDR_SIZE-1:0]			hps_read_addr;
output logic [DPR_ADDR_SIZE-1:0]			hps_next_addr;
output logic 									hps_can_read;

output logic [DPR_ADDR_SIZE-1:0]     	data_write_addr;
		 logic [DPR_ADDR_SIZE-1:0]     	data_write_addr_next;
		 logic                         	sample_can_write;
		 
output logic 									triggered;
output logic [1:0]							triggered_block_id;
output logic [28:0]							triggered_address;
       logic [3:0]							trig_outs;

		 
// HPS Ring Buffer Manager
ring_buffer_manager hps_buf_manager(
	.ring_buffer_size				(DPR_NUM_ENTRIES),
	
	.current_write_pointer		(data_write_addr), 
	.next_write_pointer			(data_write_addr_next),
	.can_write						(sample_can_write), 

	.current_read_pointer		(hps_read_addr), 
	.next_read_pointer			(hps_next_addr),
	.can_read						(hps_can_read)
);

// Arbiter between the blocks

logic [3:0] arb_reqs;
logic [3:0] arb_grants;

vc_RoundRobinArb arbtier(
	.clk			(clk),
	.reset		(reset),
	.reqs			(arb_reqs),
	.grants		(arb_grants)
);


genvar x;
generate

	for (x = 0; x < 4; x++) begin: gen_ablocks
		
		// Assign sampler Id
		logic [1:0] sampler_id;
		assign sampler_id = x;
		
		logic block_can_read;
 
		
		analyzer_block ablock(
			.clk						(clk), 
			.reset					(reset), 
			
			.inputs					(inputs[3+4*x:0+4*x]),  // [0:3], [4:7], [8:11], [12:15]
			.configuration			(block_configs[x]), 
			.trig_config			(trig_configs[x]),
			.trig_cond				(trig_conds),
			.trig_out				(trig_outs[x]),
			
			.sampler_id   			(sampler_id),
			
			.data_read_data		(block_data_out[x]), 
			.data_read_addr		(block_addr[x]), 
			.data_read_next_addr	(block_next_addr[x]), 
			.data_can_read			(block_can_read)
		);
		
		always_comb begin
			arb_reqs[x] = block_can_read;
		end
		 
	end

endgenerate

// Handle writing to the HPS SRAM
integer i;
always @(posedge clk) begin
	for (i = 0; i < 4; i++) begin
		if (reset) begin
			data_write_addr <= 1;
			block_addr[i]	 <= 0; // read from analyzer block
			data_out			 <= 0;
		end else if (arb_grants[i]) begin
			if (sample_can_write || !triggered) begin
				data_write_addr <= data_write_addr_next; // advance write address
				block_addr[i] 	 <= block_next_addr[i];	 // advance read address
				data_out        <= block_data_out[i];
			end
		end
		
	end
end

// Handle triggering
always @(posedge clk) begin
	if (reset) begin
		triggered 			 <= 0;
		triggered_block_id <= 0;
		triggered_address  <= 0;
	end else begin
		if (!triggered && (trig_outs[0] || trig_outs[1] || trig_outs[2] || trig_outs[3])) begin
			triggered 			<= 1;
			triggered_address <= data_write_addr;
			
			if 	  (trig_outs[0]) triggered_block_id <= 2'd0;
			else if (trig_outs[1]) triggered_block_id <= 2'd1;
			else if (trig_outs[2]) triggered_block_id <= 2'd2;
			else if (trig_outs[3]) triggered_block_id <= 2'd3;
		end
	end
end



// Data output is valid if some grant was given AND we can write to the HPS SRAM
assign data_out_val = arb_grants != 4'd0 && sample_can_write;


endmodule